from __future__ import annotations

import math
import os
import pickle


def test_run(dials_regression):
    from dials.algorithms.background.simple import (
        Linear2dModeller,
        MosflmOutlierRejector,
    )

    # The directory path
    path = os.path.join(
        dials_regression, "integration_test_data", "i04-weak-data", "jmp_mosflm_test"
    )

    # The input files
    reflection_filename = os.path.join(path, "mosflm_reflections.pickle")
    shoebox_filename = os.path.join(path, "shoeboxes.pickle")
    fraction = 1.0
    n_sigma = 4
    outlier_rejector = MosflmOutlierRejector(fraction, n_sigma)
    linear_modeller = Linear2dModeller()

    from dials.algorithms.shoebox import MaskCode
    from dials.array_family import flex

    print(shoebox_filename)
    # Read the data
    rtable = flex.reflection_table.from_file(reflection_filename)
    with open(shoebox_filename, "rb") as fh:
        shoeboxes, masks = pickle.load(fh, encoding="bytes")
    assert len(rtable) == len(shoeboxes)
    assert len(rtable) == len(masks)

    # Compute the background for each reflection and check against the values
    # read from the mosflm.lp file. Currently this fails for 1 strange
    # reflection whose pixel values in the mosflm file do not match those
    # extracted from the images.
    errors = 0
    VAR1 = []
    VAR2 = []
    DIFF = []
    for i in range(len(rtable)):
        I = rtable[i]["intensity.sum.value"]
        Ivar = rtable[i]["intensity.sum.variance"]
        data = shoeboxes[i].as_double()
        mask = masks[i]
        assert len(data.all()) == 2
        assert len(mask.all()) == 2
        data.reshape(flex.grid(1, *data.all()))
        mask.reshape(flex.grid(1, *mask.all()))
        outlier_rejector(data, mask)
        mask2 = (mask.as_1d() & int(MaskCode.BackgroundUsed)) != 0
        mask2.reshape(flex.grid(*mask.all()))
        model = linear_modeller.create(data, mask2)
        assert len(model.params()) == 3
        hy = data.all()[1] // 2
        hx = data.all()[2] // 2
        c1 = model.params()[0]
        a1 = model.params()[1]
        b1 = model.params()[2]
        c3 = c1 + a1 * (0.5 + hx) + b1 * (0.5 + hy)
        a2 = rtable[i]["background"][0]
        b2 = rtable[i]["background"][1]
        c2 = rtable[i]["background"][2]

        try:
            assert abs(a1 - b2) < 0.01
            assert abs(b1 + a2) < 0.01
            assert abs(c3 - c2) < 0.1
        except Exception:
            errors += 1
            continue

        background = data.as_double()
        # hy = background.all()[1] // 2
        # hx = background.all()[2] // 2
        for jj in range(background.all()[1]):
            for ii in range(background.all()[2]):
                # x = ii - hx
                # y = jj - hy
                x = ii + 0.5
                y = jj + 0.5
                background[0, jj, ii] = a1 * x + b1 * y + c1

        # Test the summation results. Edge reflections use profile fitted
        # intensity in MOSFLM. Therefore ignore these. There also appears to be a
        # some discrepancy with very low <= 0 reflections where an extra 0.5 is
        # added. Not sure why this is so ignore these reflections as well.
        from dials.algorithms.integration.sum import integrate_by_summation

        intensity = integrate_by_summation(data.as_double(), background, mask)
        I2 = intensity.intensity()
        Ivar2 = intensity.variance()
        I1 = I
        Ivar1 = Ivar
        if mask.count(0) == 0 and mask.count(2) == 0 and I1 > 0:
            VAR1.append(math.sqrt(Ivar1))
            VAR2.append(math.sqrt(Ivar2))
            DIFF.append(math.sqrt(Ivar1) - math.sqrt(Ivar2))
            try:
                assert abs(I1 - I2) < 1.0
                assert abs(math.sqrt(Ivar1) - math.sqrt(Ivar2)) < 1.0
            except Exception:
                errors += 1
                continue

    # Only 1 should fail
    assert errors == 1



# LIBTBX_SET_DISPATCHER_NAME dev.dials.show_extensions


from __future__ import annotations

import dials.util


class Script:
    """The class to encapsulate the script."""

    def __init__(self):
        """Initialise the script."""
        from libtbx.phil import parse

        from dials.util.options import ArgumentParser

        # Create the phil parameters
        phil_scope = parse(
            """
      interfaces=False
        .type = bool
        .help = "Only show information about the interfaces"
    """
        )

        # Create the option parser
        usage = "dev.dials.show_extensions [options] /path/to/image/files"
        self.parser = ArgumentParser(usage=usage, phil=phil_scope)

    def run(self, args=None):
        """Run the script."""
        import dials.extensions

        # Parse the command line arguments
        params, options = self.parser.parse_args(args)

        # Create the list of interfaces
        interfaces = [
            dials.extensions.ProfileModel,
            dials.extensions.Background,
            dials.extensions.Centroid,
            dials.extensions.SpotFinderThreshold,
        ]

        # Loop through all the interfaces
        for iface in interfaces:
            print("-" * 80)
            print(f"Extension interface: {iface.__name__}")

            # Either just show information about interfaces or show some about
            # extensions depending on user input
            if params.interfaces:

                # Print info about interface
                if options.verbose > 0:
                    print(f" name = {iface.name}")
                    if options.verbose > 1:
                        level = options.verbose - 2
                        scope = iface.phil_scope()
                        phil = scope.as_str(print_width=80 - 3, attributes_level=level)
                        phil = "\n".join((" " * 2) + l for l in phil.split("\n"))
                        if phil.strip() != "":
                            print(f" phil:\n{phil}")

            else:

                # Loop through all the extensions
                for ext in iface.extensions():
                    print(f" Extension: {ext.__name__}")
                    if options.verbose > 0:
                        print(f"  name = {ext.name}")
                        if options.verbose > 1:
                            level = options.verbose - 2
                            scope = ext.phil_scope()
                            phil = scope.as_str(
                                print_width=80 - 3, attributes_level=level
                            )
                            phil = "\n".join((" " * 3) + l for l in phil.split("\n"))
                            if phil.strip() != "":
                                print(f"  phil:\n{phil}")


@dials.util.show_mail_handle_errors()
def run(args=None):
    script = Script()
    script.run(args)


if __name__ == "__main__":
    run()



"""Tests for ΔCC½ algorithms."""

from __future__ import annotations

from unittest import mock

from dxtbx.model import Crystal, Experiment, ExperimentList, Scan

from dials.algorithms.statistics.cc_half_algorithm import CCHalfFromDials
from dials.array_family import flex
from dials.command_line.compute_delta_cchalf import phil_scope


def generated_exp(n=1):
    """Generate an experiment list with two experiments."""
    experiments = ExperimentList()
    exp_dict = {
        "__id__": "crystal",
        "real_space_a": [1.0, 0.0, 0.0],
        "real_space_b": [0.0, 1.0, 0.0],
        "real_space_c": [0.0, 0.0, 2.0],
        "space_group_hall_symbol": " C 2y",
    }
    for i in range(n):
        experiments.append(
            Experiment(
                scan=Scan(image_range=[1, 25], oscillation=[0.0, 1.0]),
                crystal=Crystal.from_dict(exp_dict),
                identifier=str(i),
            )
        )
    return experiments


def generated_refl():
    """Generate test data."""
    refls = flex.reflection_table()
    refls["intensity.scale.value"] = flex.double(range(50))
    refls["intensity.scale.variance"] = flex.double(range(50))
    refls["inverse_scale_factor"] = flex.double(50, 1.0)
    refls["id"] = flex.int([0] * 25 + [1] * 25)
    refls["xyzobs.px.value"] = flex.vec3_double(
        [(0, 0, i + 0.5) for i in range(25)] * 2
    )
    vals = [0, 1, 2, 3, 48, 49]
    # set first two of first sweep, and last two of last sweep as outliers
    outliers = flex.bool(50, False)
    for v in vals:
        outliers[v] = True
    refls.set_flags(outliers, refls.flags.outlier_in_scaling)
    refls.experiment_identifiers()[0] = "0"
    refls.experiment_identifiers()[1] = "1"
    return refls


def test_setup_of_CCHalfFromDials():
    """Test the correct setup in image group mode.

    Test for the case of outliers at the end of images, and image ranges not
    equaling a multiple of the grouping."""

    params = phil_scope.extract()
    params.mode = "image_group"
    expts = generated_exp(n=2)
    refls = generated_refl()

    # Expected behaviour is that the outliers will not be included in the
    # image range, and that all groups will have at least 10 images in.
    script = CCHalfFromDials(params, expts, refls)
    assert script.group_to_datasetid_and_range == {
        0: ("0", (5, 14)),
        1: ("0", (15, 25)),
        2: ("1", (1, 10)),
        3: ("1", (11, 23)),
    }
    assert script.datasetid_to_groups == {"0": [0, 1], "1": [2, 3]}


def test_exclusion_in_CCHalfFromDials():
    """Test the exclusion of image groups."""
    # Same input as above, but mock DeltaCCHalf algorithm to just test
    # interpretation of results and setting of excluded regions. With the
    # input, test that outlier edges are correctly removed.

    params = phil_scope.extract()
    params.mode = "image_group"
    expts = generated_exp(n=2)
    refls = generated_refl()

    def mock_algorithm(*_):
        """Mock a result from DeltaCCHalf"""
        algo = mock.Mock()
        algo.run.return_value = None
        algo.results_summary = {
            "per_dataset_delta_cc_half_values": {
                "delta_cc_half_values": [-5.0, -2.0, 4.0, -5.0],
                "datasets": [0, 1, 2, 3],
            },
            "dataset_removal": {"cutoff_value": -1.0},
        }
        return algo

    with mock.patch(
        "dials.algorithms.statistics.cc_half_algorithm.DeltaCCHalf",
        side_effect=mock_algorithm,
    ):
        script = CCHalfFromDials(params, expts, refls)
        script.run()

        assert script.datasetid_to_groups == {"0": [], "1": [2]}  # all but 3 removed
        expts = script.experiments

        assert list(expts.identifiers()) == ["1"]
        assert expts[0].scan.get_valid_image_ranges(expts.identifiers()[0]) == [(1, 10)]
        assert script.results_summary["dataset_removal"][
            "experiment_ids_fully_removed"
        ] == [0]
        assert script.results_summary["dataset_removal"][
            "experiments_fully_removed"
        ] == ["0"]



from __future__ import annotations

import os
import sys

import numpy as np

import iotbx.phil
from cctbx import uctbx
from dxtbx.model.experiment_list import ExperimentListFactory
from scitbx.math import five_number_summary

import dials.util
from dials.array_family import flex
from dials.util import Sorry, tabulate

help_message = """

Examples::

  dials.show models.expt

  dials.show image_*.cbf

  dials.show observations.refl
"""

phil_scope = iotbx.phil.parse(
    """\
show_scan_varying = False
  .type = bool
  .help = "Whether or not to show the crystal at each scan point."
show_shared_models = False
  .type = bool
  .help = "Show which models are linked to which experiments"
show_all_reflection_data = False
  .type = bool
  .help = "Whether or not to print individual reflections"
show_intensities = False
  .type = bool
show_centroids = False
  .type = bool
show_profile_fit = False
  .type = bool
show_flags = False
  .type = bool
  .help = "Show a summary table of reflection flags"
show_identifiers = False
  .type = bool
  .help = "Show experiment identifiers map if set"
image_statistics{
  show_corrected = False
    .type = bool
    .help = "Show statistics on the distribution of values in each corrected image"
  show_raw = False
    .type = bool
    .help = "Show statistics on the distribution of values in each raw image"
}
max_reflections = None
  .type = int
  .help = "Limit the number of reflections in the output."
""",
    process_includes=True,
)


def beam_centre_mm(detector, s0):
    x, y = (None, None)
    for panel_id, panel in enumerate(detector):
        try:
            x, y = panel.get_ray_intersection(s0)
        except RuntimeError:
            continue
        else:
            if panel.is_coord_valid_mm((x, y)):
                break
            else:
                x, y = (None, None)
    return panel_id, (x, y)


def beam_centre_raw_image_px(detector, s0):
    panel_id, (x, y) = beam_centre_mm(detector, s0)
    panel = detector[panel_id]
    x_px, y_px = panel.millimeter_to_pixel((x, y))
    offset = panel.get_raw_image_offset()
    return x_px + offset[0], y_px + offset[1]


def show_beam(detector, beam):

    # standard static beam model string
    s = str(beam)

    # report whether the beam is scan-varying
    if beam.num_scan_points > 0:
        s += "    s0 sampled at " + str(beam.num_scan_points) + " scan points\n"

    # add static model beam centres
    panel_id, (x, y) = beam_centre_mm(detector, beam.get_s0())
    if panel_id >= 0 and x is not None and y is not None:
        x_px, y_px = detector[panel_id].millimeter_to_pixel((x, y))
        if len(detector) > 1:
            beam_centre_mm_str = "    mm: panel %i, (%.2f,%.2f)" % (panel_id, x, y)
            beam_centre_px_str = "    px: panel %i, (%.2f,%.2f)" % (
                panel_id,
                x_px,
                y_px,
            )
            x_raw_px, y_raw_px = beam_centre_raw_image_px(detector, beam.get_s0())
            beam_centre_raw_px_str = "    px, raw image: ({:.2f},{:.2f})".format(
                x_raw_px,
                y_raw_px,
            )
            x_raw_mm, y_raw_mm = detector[panel_id].pixel_to_millimeter(
                (x_raw_px, y_raw_px)
            )
            beam_centre_raw_mm_str = "    mm, raw image: ({:.2f},{:.2f})".format(
                x_raw_mm,
                y_raw_mm,
            )
        else:
            beam_centre_mm_str = f"    mm: ({x:.2f},{y:.2f})"
            beam_centre_px_str = f"    px: ({x_px:.2f},{y_px:.2f})"
            beam_centre_raw_px_str = ""
            beam_centre_raw_mm_str = ""

        s += "\nBeam centre: \n"
        s += beam_centre_mm_str + "\n" + beam_centre_px_str + "\n"
        if beam_centre_raw_mm_str:
            s += beam_centre_raw_mm_str + "\n"
        if beam_centre_raw_px_str:
            s += beam_centre_raw_px_str + "\n"

    # report range of scan-varying model beam centres
    if beam.num_scan_points > 0:
        # get scan-varying beam centres, ensuring all on same panel
        sv_s0 = beam.get_s0_at_scan_points()
        impacts = [beam_centre_mm(detector, s0) for s0 in sv_s0]
        pnl, xy = zip(*impacts)
        uniq_pnls = set(pnl)
        if len(uniq_pnls) > 1 or min(uniq_pnls) < 0:
            return s
        if any(e == (None, None) for e in xy):
            return s
        pnl = list(uniq_pnls)[0]
        x_mm, y_mm = zip(*xy)

        # convert to pixels
        xy = [detector[pnl].millimeter_to_pixel(e) for e in xy]
        x_px, y_px = zip(*xy)

        s += "Beam centre range (mm): ([{:.2f},{:.2f}],[{:.2f},{:.2f}])\n".format(
            min(x_mm),
            max(x_mm),
            min(y_mm),
            max(y_mm),
        )
        s += "Beam centre range (px): ([{:.2f},{:.2f}],[{:.2f},{:.2f}])\n".format(
            min(x_px),
            max(x_px),
            min(y_px),
            max(y_px),
        )

    return s


def show_goniometer(goniometer):

    # standard static goniometer model string
    s = str(goniometer)

    # report whether the goniometer is scan-varying
    if goniometer.num_scan_points > 0:
        s += (
            "    Setting rotation sampled at "
            + str(goniometer.num_scan_points)
            + " scan points\n"
        )

    return s


@dials.util.show_mail_handle_errors()
def run(args=None):
    import dials.util.log

    dials.util.log.print_banner()

    from dials.util.options import (
        ArgumentParser,
        reflections_and_experiments_from_files,
    )

    usage = "dials.show [options] models.expt | image_*.cbf"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        read_experiments_from_images=True,
        read_reflections=True,
        check_format=False,
        epilog=help_message,
    )

    params, options = parser.parse_args(args=args, show_diff_phil=True)
    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    if len(experiments) == 0 and len(reflections) == 0:
        parser.print_help()
        exit()

    if len(experiments):
        if not all(e.detector for e in experiments):
            sys.exit("Error: experiment has no detector")
        if not all(e.beam for e in experiments):
            sys.exit("Error: experiment has no beam")
        print(show_experiments(experiments, show_scan_varying=params.show_scan_varying))

        if params.image_statistics.show_raw:
            show_image_statistics(experiments, "raw")

        if params.image_statistics.show_corrected:
            show_image_statistics(experiments, "corrected")

        if params.show_shared_models:
            print()
            print(model_connectivity(experiments))

    if len(reflections):
        print(
            show_reflections(
                reflections,
                show_intensities=params.show_intensities,
                show_profile_fit=params.show_profile_fit,
                show_centroids=params.show_centroids,
                show_all_reflection_data=params.show_all_reflection_data,
                show_flags=params.show_flags,
                max_reflections=params.max_reflections,
                show_identifiers=params.show_identifiers,
            )
        )


def show_experiments(experiments, show_scan_varying=False):

    text = []
    for i_expt, expt in enumerate(experiments):
        text.append("Experiment %i:" % i_expt)
        if expt.imageset:
            format_class = expt.imageset.get_format_class()
            if not format_class.is_abstract():
                text.append(f"Format class: {format_class.__name__}")
        if expt.identifier != "":
            text.append(f"Experiment identifier: {expt.identifier}")
        # get_template() only exists for ImageSequence, not ImageSet
        try:
            text.append(f"Image template: {expt.imageset.get_template()}")
        except AttributeError:
            pass
        text.append(str(expt.detector))
        text.append(
            "Max resolution (at corners): %f"
            % (expt.detector.get_max_resolution(expt.beam.get_s0()))
        )
        text.append(
            "Max resolution (inscribed):  %f"
            % (expt.detector.get_max_inscribed_resolution(expt.beam.get_s0()))
        )
        text.append("")
        text.append(show_beam(expt.detector, expt.beam))
        if expt.scan is not None:
            text.append(str(expt.scan))
        if expt.goniometer is not None:
            text.append(show_goniometer(expt.goniometer))

        if expt.crystal is not None:
            text.append(expt.crystal.as_str(show_scan_varying=show_scan_varying))
            if expt.crystal.num_scan_points:
                abc = flex.vec3_double()
                angles = flex.vec3_double()
                for n in range(expt.crystal.num_scan_points):
                    (
                        a,
                        b,
                        c,
                        alpha,
                        beta,
                        gamma,
                    ) = expt.crystal.get_unit_cell_at_scan_point(n).parameters()
                    abc.append((a, b, c))
                    angles.append((alpha, beta, gamma))
                a, b, c = abc.mean()
                alpha, beta, gamma = angles.mean()
                mean_unit_cell = uctbx.unit_cell((a, b, c, alpha, beta, gamma))
                text.append(f"  Average unit cell: {mean_unit_cell}")
        if expt.profile is not None:
            text.append(str(expt.profile))
        if expt.scaling_model is not None:
            text.append(str(expt.scaling_model))

    return "\n".join(text)


def show_image_statistics(experiments, im_type):

    if im_type == "raw":
        raw = True
    elif im_type == "corrected":
        raw = False
    else:
        raise ValueError(f"Unknown im_type: {im_type}")

    # To show image statistics, check_format has to be true. So we have to reinstatiate
    # the experiment list here
    try:
        experiments = ExperimentListFactory.from_json(
            experiments.as_json(), check_format=True
        )
    except OSError as e:
        raise Sorry(
            f"Unable to read image data. Please check {e.filename} is accessible"
        )

    print(f"Five number summary of the {im_type} images")
    for i_expt, expt in enumerate(experiments):
        for i in range(len(expt.imageset)):
            identifier = os.path.basename(expt.imageset.get_image_identifier(i))
            if raw:
                pnl_data = expt.imageset.get_raw_data(i)
            else:
                pnl_data = expt.imageset.get_corrected_data(i)
            if not isinstance(pnl_data, tuple):
                pnl_data = (pnl_data,)
            flat_data = pnl_data[0].as_1d()
            for p in pnl_data[1:]:
                flat_data.extend(p.as_1d())
            fns = five_number_summary(flat_data)
            print(
                "{}: Min: {:.1f} Q1: {:.1f} Med: {:.1f} Q3: {:.1f} Max: {:.1f}".format(
                    identifier, *fns
                )
            )


def model_connectivity(experiments):
    def model_connectivity_impl(experiments, model):
        text = [""]
        text.append(f"{model.capitalize()}:")
        models = getattr(experiments, f"{model}s")()
        rows = [[""] + [str(j) for j in range(len(models))]]
        for j, e in enumerate(experiments):
            row = ["Experiment %d" % j]
            for m in models:
                if getattr(e, model) is m:
                    row.append("x")
                else:
                    row.append(".")
            rows.append(row)
        text.append(tabulate(rows, tablefmt="plain"))
        return text

    if len(experiments) == 1:
        return ""

    text = []
    text.append("Experiment / Models")
    text.extend(model_connectivity_impl(experiments, "detector"))
    text.extend(model_connectivity_impl(experiments, "crystal"))
    text.extend(model_connectivity_impl(experiments, "beam"))

    return "\n".join(text)


def _create_flag_count_table(table):
    """Generate a summary table of flag values in a reflection table.

    :param table: A reflection table
    :returns:     A string of the formatted flags table
    """

    # Calculate the counts of entries that match each flag
    numpy_flags = table["flags"].as_numpy_array()
    flag_count = {
        flag: np.sum(numpy_flags & value != 0)
        for value, flag in table.flags.values.items()
    }

    # Work out the numeric-value order of the flags
    flag_order = sorted(table.flags.values.values(), key=lambda x: x.real)

    # Build the actual table
    flag_rows = [["Flag", "Count", "%"]]
    max_count_len = max(5, len(str(max(flag_count.values()))))
    last_flag = None
    for flag in flag_order:
        indent = ""
        # As a hint for reading, indent any 'summary' flags.
        # A summary flag is any flag which overlaps with the previous one.
        if last_flag and (last_flag.real & flag.real):
            indent = "  "
        last_flag = flag
        # Add the row to the table we're building
        flag_rows.append(
            [
                indent + flag.name,
                "{:{:d}d}".format(flag_count[flag], max_count_len),
                f"{100 * flag_count[flag] / len(table):5.01f}",
            ]
        )

    # Build the array of output strings
    text = []
    text.append("Reflection flags:")
    text.append(tabulate(flag_rows, headers="firstrow"))
    return "\n".join(text)


def show_reflections(
    reflections,
    show_intensities=False,
    show_profile_fit=False,
    show_centroids=False,
    show_all_reflection_data=False,
    show_flags=False,
    max_reflections=None,
    show_identifiers=False,
):

    text = []

    from orderedset import OrderedSet

    formats = {
        "miller_index": "%i, %i, %i",
        "d": "%.2f",
        "qe": "%.3f",
        "dqe": "%.3f",
        "id": "%i",
        "imageset_id": "%i",
        "panel": "%i",
        "flags": "%i",
        "background.mean": "%.1f",
        "background.dispersion": "%.1f",
        "background.mse": "%.1f",
        "background.sum.value": "%.1f",
        "background.sum.variance": "%.1f",
        "intensity.prf.value": "%.1f",
        "intensity.prf.variance": "%.1f",
        "intensity.sum.value": "%.1f",
        "intensity.sum.variance": "%.1f",
        "intensity.cor.value": "%.1f",
        "intensity.cor.variance": "%.1f",
        "intensity.scale.value": "%.1f",
        "intensity.scale.variance": "%.1f",
        "Ih_values": "%.1f",
        "lp": "%.3f",
        "num_pixels.background": "%i",
        "num_pixels.background_used": "%i",
        "num_pixels.foreground": "%i",
        "num_pixels.valid": "%i",
        "partial_id": "%i",
        "partiality": "%.4f",
        "profile.correlation": "%.3f",
        "profile.rmsd": "%.3f",
        "xyzcal.mm": "%.2f, %.2f, %.2f",
        "xyzcal.px": "%.2f, %.2f, %.2f",
        "delpsical.rad": "%.3f",
        "delpsical2": "%.3f",
        "delpsical.weights": "%.3f",
        "xyzobs.mm.value": "%.2f, %.2f, %.2f",
        "xyzobs.mm.variance": "%.4e, %.4e, %.4e",
        "xyzobs.px.value": "%.2f, %.2f, %.2f",
        "xyzobs.px.variance": "%.4f, %.4f, %.4f",
        "s1": "%.4f, %.4f, %.4f",
        "s2": "%.4f, %.4f, %.4f",
        "shoebox": "%.1f",
        "rlp": "%.4f, %.4f, %.4f",
        "zeta": "%.3f",
        "x_resid": "%.3f",
        "x_resid2": "%.3f",
        "y_resid": "%.3f",
        "y_resid2": "%.3f",
        "kapton_absorption_correction": "%.3f",
        "kapton_absorption_correction_sigmas": "%.3f",
        "inverse_scale_factor": "%.3f",
        "inverse_scale_factor_variance": "%.3f",
    }

    for rlist in reflections:
        from dials.algorithms.shoebox import MaskCode

        foreground_valid = MaskCode.Valid | MaskCode.Foreground
        text.append("")
        text.append(f"Reflection list contains {len(rlist)} reflections")

        if len(rlist) == 0:
            continue

        rows = [["Column", "min", "max", "mean"]]
        for k, col in rlist.cols():
            if k in formats and "%" not in formats.get(k, "%s"):
                # Allow blanking out of entries that wouldn't make sense
                rows.append(
                    [
                        k,
                        formats.get(k, "%s"),
                        formats.get(k, "%s"),
                        formats.get(k, "%s"),
                    ]
                )
            elif type(col) in (flex.double, flex.int, flex.size_t):
                if type(col) in (flex.int, flex.size_t):
                    col = col.as_double()
                rows.append(
                    [
                        k,
                        formats.get(k, "%s") % flex.min(col),
                        formats.get(k, "%s") % flex.max(col),
                        formats.get(k, "%s") % flex.mean(col),
                    ]
                )
            elif type(col) in (flex.vec3_double, flex.miller_index):
                if isinstance(col, flex.miller_index):
                    col = col.as_vec3_double()
                rows.append(
                    [
                        k,
                        formats.get(k, "%.2f, %.2f, %.2f") % col.min(),
                        formats.get(k, "%.2f, %.2f, %.2f") % col.max(),
                        formats.get(k, "%.2f, %.2f, %.2f") % col.mean(),
                    ]
                )
            elif isinstance(col, flex.shoebox):
                rows.append([k, "", "", ""])
                si = col.summed_intensity().observed_value()
                rows.append(
                    [
                        "  summed I",
                        formats.get(k, "%s") % flex.min(si),
                        formats.get(k, "%s") % flex.max(si),
                        formats.get(k, "%s") % flex.mean(si),
                    ]
                )
                x1, x2, y1, y2, z1, z2 = col.bounding_boxes().parts()
                bbox_sizes = ((z2 - z1) * (y2 - y1) * (x2 - x1)).as_double()
                rows.append(
                    [
                        "  N pix",
                        formats.get(k, "%s") % flex.min(bbox_sizes),
                        formats.get(k, "%s") % flex.max(bbox_sizes),
                        formats.get(k, "%s") % flex.mean(bbox_sizes),
                    ]
                )
                fore_valid = col.count_mask_values(foreground_valid).as_double()
                rows.append(
                    [
                        "  N valid foreground pix",
                        formats.get(k, "%s") % flex.min(fore_valid),
                        formats.get(k, "%s") % flex.max(fore_valid),
                        formats.get(k, "%s") % flex.mean(fore_valid),
                    ]
                )

        text.append(tabulate(rows, headers="firstrow"))

        if show_flags:
            text.append(_create_flag_count_table(rlist))

        if show_identifiers:
            if rlist.experiment_identifiers():
                text.append(
                    """Experiment identifiers id-map values:\n%s"""
                    % (
                        "\n".join(
                            "id:"
                            + str(k)
                            + " -> experiment identifier:"
                            + str(rlist.experiment_identifiers()[k])
                            for k in rlist.experiment_identifiers().keys()
                        )
                    )
                )

    intensity_keys = (
        "miller_index",
        "d",
        "intensity.prf.value",
        "intensity.prf.variance",
        "intensity.sum.value",
        "intensity.sum.variance",
        "background.mean",
        "profile.correlation",
        "profile.rmsd",
    )

    profile_fit_keys = ("miller_index", "d")

    centroid_keys = (
        "miller_index",
        "d",
        "xyzcal.mm",
        "xyzcal.px",
        "xyzobs.mm.value",
        "xyzobs.mm.variance",
        "xyzobs.px.value",
        "xyzobs.px.variance",
    )

    keys_to_print = OrderedSet()

    if show_intensities:
        for k in intensity_keys:
            keys_to_print.add(k)
    if show_profile_fit:
        for k in profile_fit_keys:
            keys_to_print.add(k)
    if show_centroids:
        for k in centroid_keys:
            keys_to_print.add(k)
    if show_all_reflection_data:
        for k in formats:
            keys_to_print.add(k)

    def format_column(key, data, format_strings=None):
        if isinstance(data, flex.vec3_double):
            c_strings = [
                c.as_string(format_strings[i].strip())
                for i, c in enumerate(data.parts())
            ]
        elif isinstance(data, flex.miller_index):
            c_strings = [
                c.as_string(format_strings[i].strip())
                for i, c in enumerate(data.as_vec3_double().parts())
            ]
        elif isinstance(data, flex.size_t):
            c_strings = [data.as_int().as_string(format_strings[0].strip())]
        elif isinstance(data, flex.shoebox):
            x1, x2, y1, y2, z1, z2 = data.bounding_boxes().parts()
            bbox_sizes = ((z2 - z1) * (y2 - y1) * (x2 - x1)).as_double()
            c_strings = [bbox_sizes.as_string(format_strings[0].strip())]
            key += " (N pix)"
        else:
            c_strings = [data.as_string(format_strings[0].strip())]

        column = flex.std_string()
        max_element_lengths = [c.max_element_length() for c in c_strings]
        for i in range(len(c_strings[0])):

            column.append(
                f"%{len(key)}s"
                % ", ".join(
                    ("%%%is" % max_element_lengths[j]) % c_strings[j][i]
                    for j in range(len(c_strings))
                )
            )
        return column

    if keys_to_print:
        keys = [k for k in keys_to_print if k in rlist]
        if max_reflections is not None:
            max_reflections = min(len(rlist), max_reflections)
        else:
            max_reflections = len(rlist)

        columns = []

        for k in keys:
            columns.append(
                format_column(k, rlist[k], format_strings=formats[k].split(","))
            )

        text.append("")
        text.append("Printing %i of %i reflections:" % (max_reflections, len(rlist)))
        line = []
        for j in range(len(columns)):
            key = keys[j]
            if key == "shoebox":
                key += " (N pix)"
            width = max(len(key), columns[j].max_element_length())
            line.append("%%%is" % width % key)
        text.append(" ".join(line))
        for i in range(max_reflections):
            line = (c[i] for c in columns)
            text.append(" ".join(line))

    return "\n".join(text)


if __name__ == "__main__":
    run()



from __future__ import annotations

import json
import os

import procrunner
import pytest

from cctbx import sgtbx, uctbx
from dxtbx.serialize import load

from dials.command_line import refine_bravais_settings


def test_refine_bravais_settings_i04_weak_data(dials_regression, tmp_path):
    data_dir = os.path.join(dials_regression, "indexing_test_data", "i04_weak_data")
    pickle_path = os.path.join(data_dir, "indexed.pickle")
    experiments_path = os.path.join(data_dir, "experiments.json")
    result = procrunner.run(
        [
            "dials.refine_bravais_settings",
            pickle_path,
            experiments_path,
            "reflections_per_degree=5",
            "minimum_sample_size=500",
            "beam.fix=all",
            "detector.fix=all",
            "prefix=tst_",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    for i in range(1, 10):
        assert (tmp_path / f"tst_bravais_setting_{i}.expt").is_file()

    experiments_list = load.experiment_list(
        tmp_path / "tst_bravais_setting_9.expt", check_format=False
    )
    assert len(experiments_list) == 1
    assert (
        experiments_list[0]
        .crystal.get_unit_cell()
        .is_similar_to(uctbx.unit_cell((57.782, 57.782, 150.011, 90, 90, 90)))
    )
    assert experiments_list[0].crystal.get_space_group().type().hall_symbol() == " P 4"

    assert (tmp_path / "tst_bravais_summary.json").is_file()
    with (tmp_path / "tst_bravais_summary.json").open("rb") as fh:
        bravais_summary = json.load(fh)
    assert set(bravais_summary) == {"1", "2", "3", "4", "5", "6", "7", "8", "9"}
    assert set(bravais_summary["9"]).issuperset(
        {"bravais", "max_angular_difference", "unit_cell", "rmsd", "nspots"}
    )

    assert bravais_summary["9"]["unit_cell"] == pytest.approx(
        [57.78, 57.78, 150.0, 90.0, 90.0, 90.0], abs=1e-1
    )
    assert bravais_summary["9"]["bravais"] == "tP"
    assert bravais_summary["9"]["recommended"] is True
    assert bravais_summary["9"]["rmsd"] == pytest.approx(0.047, abs=1e-2)


def test_refine_bravais_settings_multi_sweep(dials_regression, tmp_path):
    data_dir = os.path.join(dials_regression, "indexing_test_data", "multi_sweep")
    pickle_path = os.path.join(data_dir, "indexed.pickle")
    experiments_path = os.path.join(data_dir, "experiments.json")
    result = procrunner.run(
        ["dials.refine_bravais_settings", pickle_path, experiments_path],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    for i in range(1, 10):
        assert (tmp_path / f"bravais_setting_{i}.expt").is_file()

    experiments_list = load.experiment_list(
        tmp_path / "bravais_setting_9.expt", check_format=False
    )
    assert len(experiments_list) == 4
    assert len(experiments_list.crystals()) == 1
    assert (
        experiments_list[0]
        .crystal.get_unit_cell()
        .is_similar_to(uctbx.unit_cell((7.31, 7.31, 6.82, 90.00, 90.00, 90.00)))
    )
    assert experiments_list[0].crystal.get_space_group().type().hall_symbol() == " I 4"
    assert (tmp_path / "bravais_summary.json").is_file()
    with (tmp_path / "bravais_summary.json").open("rb") as fh:
        bravais_summary = json.load(fh)
    for i in range(1, 23):
        assert str(i) in bravais_summary

    assert bravais_summary["9"]["unit_cell"] == pytest.approx(
        [7.31, 7.31, 6.82, 90.00, 90.00, 90.00], abs=1e-1
    )
    assert bravais_summary["9"]["bravais"] == "tI"
    assert bravais_summary["9"]["rmsd"] == pytest.approx(0.103, abs=1e-2)
    assert bravais_summary["9"]["recommended"] is True


def test_refine_bravais_settings_trypsin(dials_regression, tmp_path):
    data_dir = os.path.join(dials_regression, "indexing_test_data", "trypsin")
    pickle_path = os.path.join(data_dir, "indexed.pickle")
    experiments_path = os.path.join(data_dir, "experiments.json")
    result = procrunner.run(
        [
            "dials.refine_bravais_settings",
            pickle_path,
            experiments_path,
            "crystal_id=1",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    for i in range(1, 10):
        assert (tmp_path / f"bravais_setting_{i}.expt").is_file()

    experiments_list = load.experiment_list(
        tmp_path / "bravais_setting_5.expt", check_format=False
    )
    assert len(experiments_list) == 1
    assert (
        experiments_list[0]
        .crystal.get_unit_cell()
        .is_similar_to(uctbx.unit_cell((54.37, 58.29, 66.51, 90.00, 90.00, 90.00)))
    )
    assert (
        experiments_list[0].crystal.get_space_group().type().hall_symbol() == " P 2 2"
    )

    assert (tmp_path / "bravais_summary.json").is_file()
    with (tmp_path / "bravais_summary.json").open("rb") as fh:
        bravais_summary = json.load(fh)
    assert set(bravais_summary) == {"1", "2", "3", "4", "5", "6", "7", "8", "9"}

    assert bravais_summary["5"]["unit_cell"] == pytest.approx(
        [54.37, 58.29, 66.51, 90.00, 90.00, 90.00], abs=1e-1
    )
    assert bravais_summary["5"]["bravais"] == "oP"
    assert bravais_summary["5"]["rmsd"] == pytest.approx(0.1200, abs=1e-2)
    assert bravais_summary["5"]["recommended"] is True
    assert bravais_summary["9"]["recommended"] is False


def test_refine_bravais_settings_554(dials_regression, tmp_path):
    data_dir = os.path.join(dials_regression, "dials-554")
    pickle_path = os.path.join(data_dir, "indexed.pickle")
    experiments_path = os.path.join(data_dir, "experiments.json")
    result = procrunner.run(
        ["dials.refine_bravais_settings", pickle_path, experiments_path],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    for i in range(1, 5):
        assert (tmp_path / f"bravais_setting_{i}.expt").is_file()

    experiments_list = load.experiment_list(
        tmp_path / "bravais_setting_5.expt", check_format=False
    )
    assert len(experiments_list) == 7
    assert len(experiments_list.crystals()) == 1
    crystal = experiments_list.crystals()[0]
    assert crystal.get_unit_cell().is_similar_to(
        uctbx.unit_cell((4.75863, 4.75863, 12.9885, 90, 90, 120))
    )
    assert crystal.get_space_group().type().hall_symbol() == " R 3"
    # assert all of the detectors are different
    for expt in experiments_list[1:]:
        assert expt.detector != experiments_list[0].detector
    for i in (0, 1, 6):
        assert experiments_list[i].detector[0].get_origin() == pytest.approx(
            (-41, 5.5, -135), abs=1
        )
    for i in (2, 3, 4, 5):
        assert experiments_list[i].detector[0].get_origin() == pytest.approx(
            (-41, 91, -99), abs=1
        )
    assert (tmp_path / "bravais_summary.json").is_file()
    with (tmp_path / "bravais_summary.json").open("rb") as fh:
        bravais_summary = json.load(fh)
    for i in range(1, 5):
        assert str(i) in bravais_summary

    assert bravais_summary["5"]["unit_cell"] == pytest.approx(
        [4.75863, 4.75863, 12.9885, 90, 90, 120], abs=1e-1
    )
    assert bravais_summary["5"]["bravais"] == "hR"
    assert bravais_summary["5"]["rmsd"] == pytest.approx(0.104, abs=1e-2)
    assert bravais_summary["5"]["recommended"] is True


@pytest.mark.parametrize(
    "best_monoclinic_beta,expected_space_group,expected_unit_cell",
    [
        (True, "I 1 2 1", (44.47, 52.85, 111.46, 90.00, 99.91, 90.00)),
        (False, "C 1 2 1", (112.67, 52.85, 44.47, 90.00, 102.97, 90.00)),
    ],
)
def test_setting_c2_vs_i2(
    best_monoclinic_beta,
    expected_space_group,
    expected_unit_cell,
    dials_data,
    tmp_path,
):
    data_dir = dials_data("mpro_x0305_processed", pathlib=True)
    refl_path = data_dir / "indexed.refl"
    experiments_path = data_dir / "indexed.expt"
    result = procrunner.run(
        [
            "dials.refine_bravais_settings",
            os.fspath(experiments_path),
            os.fspath(refl_path),
            f"best_monoclinic_beta={best_monoclinic_beta}",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    expts_orig = load.experiment_list(experiments_path, check_format=False)
    expts = load.experiment_list(
        tmp_path / "bravais_setting_2.expt", check_format=False
    )
    expected_bravais_lattice = str(
        sgtbx.bravais_types.bravais_lattice(symbol=expected_space_group)
    )
    expts[0].crystal.get_space_group().type().lookup_symbol() == expected_space_group
    assert expts[0].crystal.get_unit_cell().parameters() == pytest.approx(
        expected_unit_cell, abs=1e-2
    )
    with (tmp_path / "bravais_summary.json").open("rb") as fh:
        bravais_summary = json.load(fh)
    assert bravais_summary["2"]["bravais"] == expected_bravais_lattice
    # Verify that the cb_op converts from the input setting to the refined setting
    cb_op = sgtbx.change_of_basis_op(str(bravais_summary["2"]["cb_op"]))
    assert (
        expts_orig[0]
        .crystal.change_basis(cb_op)
        .get_unit_cell()
        .is_similar_to(
            expts[0].crystal.get_unit_cell(),
            relative_length_tolerance=0.1,
            absolute_angle_tolerance=1,
        )
    )
    stdout = result.stdout.decode()
    assert bravais_summary["2"]["cb_op"] in stdout
    assert f"| {expected_bravais_lattice}        |" in stdout
    expected_short_name = refine_bravais_settings.short_space_group_name(
        sgtbx.space_group_info(expected_space_group).group()
    )
    assert f"{expected_bravais_lattice}: {expected_short_name}" in stdout


def test_refine_bravais_settings_non_primitive_input(dials_data, tmp_path):
    data_dir = dials_data("insulin_processed", pathlib=True)
    refl_path = data_dir / "indexed.refl"
    expt_path = data_dir / "indexed.expt"
    result = procrunner.run(
        [
            "dials.refine_bravais_settings",
            expt_path,
            refl_path,
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    for i in range(1, 22):
        assert (tmp_path / f"bravais_setting_{i}.expt").is_file()

    assert (tmp_path / "bravais_summary.json").is_file()
    with (tmp_path / "bravais_summary.json").open("rb") as fh:
        bravais_summary = json.load(fh)
    assert bravais_summary["1"]["bravais"] == "aP"
    assert bravais_summary["1"]["cb_op"] == "y+z,x+z,x+y"
    assert bravais_summary["22"]["bravais"] == "cI"
    assert bravais_summary["22"]["cb_op"] == "a,b,c"

    for record in result.stdout.decode().split("\n"):
        if "aP " in record:
            assert "y+z,x+z,x+y" in record
        elif "cI " in record:
            assert "a,b,c" in record

    # Verify that the reported cb_ops correctly map the input setting
    # to the unit cell for each Bravais setting
    input_expts = load.experiment_list(expt_path, check_format=False)
    input_cs = input_expts[0].crystal.get_crystal_symmetry()
    for i in range(22):
        uc_input_to_ref = input_cs.unit_cell().change_basis(
            sgtbx.change_of_basis_op(bravais_summary[f"{i+1}"]["cb_op"])
        )
        uc_ref = uctbx.unit_cell(bravais_summary[f"{i+1}"]["unit_cell"])
        assert uc_input_to_ref.is_similar_to(uc_ref)



"""Definitions of functions and classes for scaling and filtering algorithm."""

from __future__ import annotations

import math

from libtbx import phil
from scitbx.array_family import flex

from dials.report.plots import d_star_sq_to_d_ticks

phil_scope = phil.parse(
    """
filtering {
    method = None deltacchalf
        .type = choice
        .help = "Choice of whether to do any filtering cycles, default None."
    deltacchalf {
        max_cycles = 6
            .type = int(value_min=1)
        max_percent_removed = 10
            .type = float
        min_completeness = None
            .type = float(value_min=0, value_max=100)
            .help = "Desired minimum completeness, as a percentage (0 - 100)."
        mode = *dataset image_group
            .type = choice
            .help = "Perform analysis on whole datasets or batch groups"
        group_size = 10
            .type = int(value_min=1)
            .help = "The number of images to group together when calculating delta"
                    "cchalf in image_group mode"
        stdcutoff = 4.0
            .type = float
            .help = "Datasets with a ΔCC½ below (mean - stdcutoff*std) are removed"
    }
    output {
        scale_and_filter_results = "scale_and_filter_results.json"
            .type = str
            .help = "Filename for output json of scale and filter results."
    }
}
"""
)


def ordinal(n):
    return "%d%s" % (n, "tsnrhtdd"[(n / 10 % 10 != 1) * (n % 10 < 4) * n % 10 :: 4])


def log_cycle_results(results, scaling_script, filter_script):
    """Log results from the scripts for this cycle and add to the results."""
    cycle_results = {"merging_stats": scaling_script.merging_statistics_result}

    if not results.get_cycle_results():
        results.initial_n_reflections = scaling_script.scaled_miller_array.size()

    cycle_results["delta_cc_half_values"] = filter_script.results_summary[
        "per_dataset_delta_cc_half_values"
    ]["delta_cc_half_values"]
    cycle_results["mean_cc_half"] = filter_script.results_summary["mean_cc_half"]
    removal_summary = filter_script.results_summary["dataset_removal"]
    if removal_summary["mode"] == "image_group":
        cycle_results["image_ranges_removed"] = removal_summary["image_ranges_removed"]
    cycle_results["removed_datasets"] = removal_summary["experiments_fully_removed"]
    cycle_results["removed_ids"] = removal_summary["experiment_ids_fully_removed"]

    cycle_results["n_removed"] = filter_script.results_summary["dataset_removal"][
        "n_reflections_removed"
    ]

    n_removed = (
        sum(res["n_removed"] for res in results.get_cycle_results())
        + cycle_results["n_removed"]
    )
    percent_removed = n_removed / results.initial_n_reflections * 100
    cycle_results["cumul_percent_removed"] = percent_removed

    results.add_cycle_results(cycle_results)
    return results


class AnalysisResults:
    """Class to store results from scaling and filtering."""

    def __init__(self):
        self.termination_reason = None
        self.cycle_results = []
        self.initial_n_reflections = None
        self.initial_expids_and_image_ranges = None
        self.expids_and_image_ranges = None
        self.final_stats = None

    def add_cycle_results(self, results_dict):
        """Add the results dict from a scale and filter cycle."""
        merging_stats_dict = self._parse_merging_stats(results_dict["merging_stats"])
        results_dict["merging_stats"] = merging_stats_dict
        self.cycle_results.append(results_dict)

    @staticmethod
    def _parse_merging_stats(merging_stats_obj):
        merging_stats = {}
        overall = merging_stats_obj.overall
        merging_stats["ccs"] = [b.cc_one_half for b in merging_stats_obj.bins]
        merging_stats["rmerge"] = [b.r_merge for b in merging_stats_obj.bins]
        merging_stats["rpim"] = [b.r_pim for b in merging_stats_obj.bins]
        merging_stats["d_min"] = [b.d_min for b in merging_stats_obj.bins]
        merging_stats["overall"] = {
            "cc_one_half": overall.cc_one_half,
            "r_merge": overall.r_merge,
            "r_pim": overall.r_pim,
            "i_over_sigma_mean": overall.i_over_sigma_mean,
            "completeness": 100 * overall.completeness,
            "n_obs": overall.n_obs,
        }
        return merging_stats

    def get_cycle_results(self):
        """Get the results from all cycles."""
        return self.cycle_results

    def get_last_cycle_results(self):
        """Get the results from the latest recorded cycle."""
        return self.cycle_results[-1]

    def add_final_stats(self, final_stats):
        """Add additional final merging stats from final rescale."""
        self.final_stats = self._parse_merging_stats(final_stats)

    def get_merging_stats(self):
        """Get all merging stats, including additional final stats if present."""
        stats = [res["merging_stats"] for res in self.cycle_results]
        if self.final_stats:
            stats += [self.final_stats]
        return stats

    def finish(self, termination_reason):
        """Set the termination reason/"""
        assert termination_reason in [
            "no_more_removed",
            "max_cycles",
            "max_percent_removed",
            "below_completeness_limit",
        ]
        self.termination_reason = termination_reason

    def to_dict(self):
        """Return the stored data as a dictionary."""
        return {
            "termination_reason": self.termination_reason,
            "initial_n_reflections": self.initial_n_reflections,
            "initial_expids_and_image_ranges": self.initial_expids_and_image_ranges,
            "expids_and_image_ranges": self.expids_and_image_ranges,
            "cycle_results": {i + 1: val for i, val in enumerate(self.cycle_results)},
            "final_stats": self.final_stats,
        }

    @staticmethod
    def from_dict(dictionary):
        """Configure the class from its dictionary form."""
        results = AnalysisResults()
        results.termination_reason = dictionary["termination_reason"]
        results.initial_expids_and_image_ranges = dictionary[
            "initial_expids_and_image_ranges"
        ]
        results.expids_and_image_ranges = dictionary["expids_and_image_ranges"]
        results.cycle_results = [
            dictionary["cycle_results"][str(key)]
            for key in sorted(int(k) for k in dictionary["cycle_results"].keys())
        ]
        results.initial_n_reflections = dictionary["initial_n_reflections"]
        results.final_stats = dictionary["final_stats"]
        return results

    def __str__(self):
        """Make summary of results."""
        msg = "\nSummary of data removed:\n"
        for i, res in enumerate(self.get_cycle_results()):
            msg += f"Cycle number: {i + 1}\n"
            if "image_ranges_removed" in res:
                if res["image_ranges_removed"]:
                    removed = "\n    ".join(
                        str(t[0]) + ", dataset " + str(t[1])
                        for t in res["image_ranges_removed"]
                    )
                    msg += f"  Removed image ranges: \n    {removed}"
            else:
                if res["removed_ids"]:
                    msg += f"  Removed datasets: {res['removed_ids']}\n"
            msg += (
                "  cumulative %% of reflections removed: %.3f\n"
                % res["cumul_percent_removed"]
            )
        return msg


color_list = [
    "#F44336",
    "#FFC107",
    "#FFEB3B",
    "#8BC34A",
    "#03A9F4",
    "#3F51B5",
    "#607D8B",
]


def make_scaling_filtering_plots(data):
    """Make the plots for scaling and filtering."""
    d = make_filtering_merging_stats_plots(data["merging_stats"])
    d.update(make_histogram_plots(data["cycle_results"]))
    if data["mode"] == "image_group":
        d.update(
            make_reduction_plots(
                data["initial_expids_and_image_ranges"], data["expids_and_image_ranges"]
            )
        )
    return d


def make_filtering_merging_stats_plots(merging_stats):
    """Generate plotting dicts for merging statistics."""
    n_datasets = len(merging_stats)
    colors = [
        (color_list * int(math.ceil(n_datasets / len(color_list))))[i]
        for i in range(n_datasets)
    ]
    colors[-1] = "k"

    legends = ["initial scale"]
    if len(merging_stats) > 2:
        legends += [
            ordinal(i) + " rescale" for i in range(1, len(merging_stats) - 1)
        ] + ["final rescale"]
    elif len(merging_stats) == 2:
        legends += ["final rescale"]

    overall_ccs = [m["overall"]["cc_one_half"] for m in merging_stats]
    overall_rpim = [m["overall"]["r_pim"] for m in merging_stats]
    overall_ioversigma = [m["overall"]["i_over_sigma_mean"] for m in merging_stats]
    overall_completeness = [m["overall"]["completeness"] for m in merging_stats]
    # first make overall plots
    d = {
        "cc_one_half_vs_cycle": {
            "data": [
                {
                    "y": overall_ccs,
                    "x": list(range(1, n_datasets + 1)),
                    "type": "scatter",
                    "mode": "lines",
                }
            ],
            "layout": {
                "title": "CC<sub>½</sub> vs cycle",
                "xaxis": {"title": "Cycle number"},
                "yaxis": {"title": "CC<sub>½</sub>"},
            },
        },
        "r_pim_vs_cycle": {
            "data": [
                {
                    "y": overall_rpim,
                    "x": list(range(1, n_datasets + 1)),
                    "type": "scatter",
                    "mode": "lines",
                }
            ],
            "layout": {
                "title": "R-pim vs cycle",
                "xaxis": {"title": "Cycle number"},
                "yaxis": {"title": "R-pim"},
            },
        },
        "i_over_sigma_vs_cycle": {
            "data": [
                {
                    "y": overall_ioversigma,
                    "x": list(range(1, n_datasets + 1)),
                    "type": "scatter",
                    "mode": "lines",
                }
            ],
            "layout": {
                "title": "<I/σ(I)> vs cycle",
                "xaxis": {"title": "Cycle number"},
                "yaxis": {"title": "<I/σ(I)>"},
            },
        },
        "completeness_vs_cycle": {
            "data": [
                {
                    "y": overall_completeness,
                    "x": list(range(1, n_datasets + 1)),
                    "type": "scatter",
                    "mode": "lines",
                }
            ],
            "layout": {
                "title": "Completeness vs cycle",
                "xaxis": {"title": "Cycle number"},
                "yaxis": {"title": "Completeness"},
            },
        },
    }
    cc_one_half_bins = merging_stats[0]["ccs"]
    r_pim_bins = merging_stats[0]["rpim"]
    r_merge_bins = merging_stats[0]["rmerge"]
    resolution = [1.0 / x**2 for x in merging_stats[0]["d_min"]]
    vals, txt = d_star_sq_to_d_ticks(resolution, 5)
    d.update(
        {
            "cc_one_half_filter": {
                "data": [
                    {
                        "x": resolution,  # d_star_sq
                        "y": cc_one_half_bins,
                        "type": "scatter",
                        "name": legends[0],
                        "mode": "lines",
                        "line": {"color": colors[0]},
                    }
                ],
                "layout": {
                    "title": "CC<sub>½</sub> vs resolution",
                    "xaxis": {
                        "title": "Resolution (Å)",
                        "tickvals": vals,
                        "ticktext": txt,
                    },
                    "yaxis": {"title": "CC<sub>½</sub>", "range": [0, 1]},
                },
            }
        }
    )
    d.update(
        {
            "r_pim_filter": {
                "data": [
                    {
                        "x": resolution,  # d_star_sq
                        "y": r_pim_bins,
                        "type": "scatter",
                        "name": legends[0],
                        "mode": "lines",
                        "line": {"color": colors[0]},
                    }
                ],
                "layout": {
                    "title": "R-pim vs resolution",
                    "xaxis": {
                        "title": "Resolution (Å)",
                        "tickvals": vals,
                        "ticktext": txt,
                    },
                    "yaxis": {
                        "title": "R-pim",
                        "range": [0, min(1.5, max(r_pim_bins))],
                    },
                },
            }
        }
    )
    d.update(
        {
            "r_merge_filter": {
                "data": [
                    {
                        "x": resolution,  # d_star_sq
                        "y": r_merge_bins,
                        "type": "scatter",
                        "name": legends[0],
                        "mode": "lines",
                        "line": {"color": colors[0]},
                    }
                ],
                "layout": {
                    "title": "R-merge vs resolution",
                    "xaxis": {
                        "title": "Resolution (Å)",
                        "tickvals": vals,
                        "ticktext": txt,
                    },
                    "yaxis": {
                        "title": "R-merge",
                        "range": [0, min(1.5, max(r_merge_bins))],
                    },
                },
            }
        }
    )
    for c, stats in enumerate(merging_stats[1:]):
        cc_one_half_bins = stats["ccs"]
        r_pim_bins = stats["rpim"]
        r_merge_bins = stats["rmerge"]
        resolution = [1.0 / x**2 for x in stats["d_min"]]
        d["cc_one_half_filter"]["data"].append(
            {
                "x": resolution,  # d_star_sq
                "y": cc_one_half_bins,
                "type": "scatter",
                "name": legends[c + 1],
                "line": {"color": colors[c + 1]},
            }
        )
        d["r_pim_filter"]["data"].append(
            {
                "x": resolution,  # d_star_sq
                "y": r_pim_bins,
                "type": "scatter",
                "name": legends[c + 1],
                "line": {"color": colors[c + 1]},
            }
        )
        d["r_merge_filter"]["data"].append(
            {
                "x": resolution,  # d_star_sq
                "y": r_merge_bins,
                "type": "scatter",
                "name": legends[c + 1],
                "line": {"color": colors[c + 1]},
            }
        )
    return d


def make_histogram_plots(cycle_results):
    """Make the histogram plots."""
    delta_cc_half_lists = [res["delta_cc_half_values"] for res in cycle_results]
    if not delta_cc_half_lists:
        return {}

    n = len(delta_cc_half_lists)

    overall_mean_ccs = [res["mean_cc_half"] for res in cycle_results]
    d = {
        "mean_cc_one_half_vs_cycle": {
            "data": [
                {
                    "y": overall_mean_ccs,
                    "x": list(range(1, n + 1)),
                    "type": "scatter",
                    "mode": "lines",
                }
            ],
            "layout": {
                "title": "Resolution-averaged CC<sub>½</sub> (σ-τ) vs cycle",
                "xaxis": {"title": "Cycle number"},
                "yaxis": {"title": "Resolution-averaged CC<sub>½</sub> (σ-τ)"},
            },
        }
    }
    colors = [(color_list * int(math.ceil(n / len(color_list))))[i] for i in range(n)]
    if n == 1:
        legends = ["ΔCC<sub>½</sub> analysis"]
    else:
        legends = [ordinal(i) + " ΔCC<sub>½</sub> analysis" for i in range(1, n + 1)]
    if "image_ranges_removed" in cycle_results[0]:
        n_rej = [len(res["image_ranges_removed"]) for res in cycle_results]
    else:
        n_rej = [len(res["removed_datasets"]) for res in cycle_results]

    def _color_bar_charts(counts, index):
        n = 0
        bar_colors = []
        n_rej_this = n_rej[index]
        for count in counts:
            if n >= n_rej_this:
                bar_colors.append(colors[index])
            else:
                bar_colors.append("k")
                n += count
        return bar_colors

    def _add_new_histogram(d, hist, index):
        d.update(
            {
                f"scale_filter_histograms_{index}": {
                    "data": [
                        {
                            "x": list(hist.slot_centers()),
                            "y": list(hist.slots()),
                            "type": "bar",
                            "name": legends[index],
                            "marker": {"color": _color_bar_charts(hist.slots(), index)},
                        }
                    ],
                    "layout": {
                        "title": f"{legends[index]}",
                        "xaxis": {"title": "ΔCC<sub>½</sub>"},
                        "yaxis": {
                            "title": "Number of datasets/groups",
                            "range": [0, min(max(hist.slots()), 50)],
                        },
                        "bargap": 0,
                    },
                }
            }
        )

    for c, deltas in enumerate(delta_cc_half_lists):
        flex_deltas = flex.double(deltas)  # convert list to flex array
        if flex_deltas.all_eq(flex_deltas[0]):
            continue
        hist = flex.histogram(
            flex_deltas * 100,
            flex.min(flex_deltas) * 100,
            flex.max(flex_deltas) * 100,
            n_slots=40,
        )
        _add_new_histogram(d, hist, c)
    return d


def make_per_dataset_plot(delta_cchalf_i):
    """Make a line plot of ΔCC½ per group."""

    return {
        "per_dataset_plot": {
            "data": [
                {
                    "y": [i * 100 for i in list(delta_cchalf_i.values())],
                    "x": list(delta_cchalf_i.keys()),
                    "type": "scatter",
                    "mode": "lines",
                }
            ],
            "layout": {
                "title": "ΔCC<sub>½</sub> vs group",
                "xaxis": {"title": "Group number"},
                "yaxis": {"title": "ΔCC<sub>½</sub>"},
            },
        }
    }


def make_reduction_plots(initial_expids_and_image_ranges, expids_and_image_ranges):
    """Make a chart showing excluded image ranges."""
    x = list(range(len(initial_expids_and_image_ranges)))
    initial_n_images = []
    initial_expids = []
    for expid_and_img in initial_expids_and_image_ranges:
        initial_n_images.append(expid_and_img[1][1] - expid_and_img[1][0] + 1)
        initial_expids.append(expid_and_img[0])
    final_n_images = [0] * len(x)
    for expid, valid in expids_and_image_ranges:
        loc = initial_expids.index(expid)
        for t in valid:
            final_n_images[loc] = t[1] - t[0] + 1
    n_removed = [i - j for (i, j) in zip(initial_n_images, final_n_images)]
    d = {
        "reduction_plot": {
            "data": [
                {
                    "x": x,
                    "y": final_n_images,
                    "type": "bar",
                    "name": "final image ranges",
                },
                {"x": x, "y": n_removed, "type": "bar", "name": "removed image ranges"},
            ],
            "layout": {
                "title": "Image range plot",
                "xaxis": {"title": "Experiment number"},
                "yaxis": {"title": "Image number"},
                "bargap": 0,
                "barmode": "stack",
            },
        }
    }
    return d



from __future__ import annotations

import procrunner

import iotbx.merging_statistics


def test_indexed_as_integrated(dials_data, tmp_path):
    data_dir = dials_data("insulin_processed", pathlib=True)
    refl = data_dir / "indexed.refl"
    expt = data_dir / "indexed.expt"

    command = [
        "dials.indexed_as_integrated",
        refl,
        expt,
        "output.reflections=output.refl",
    ]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "output.refl").is_file()

    # now make sure we can run dials.symmetry and dials.scale successfully
    sym_command = ["dials.symmetry", tmp_path / "output.refl", expt]
    result = procrunner.run(sym_command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "symmetrized.refl").is_file()
    assert (tmp_path / "symmetrized.expt").is_file()

    scale_command = [
        "dials.scale",
        tmp_path / "symmetrized.refl",
        tmp_path / "symmetrized.expt",
        "unmerged_mtz=scaled.mtz",
        "d_min=2.0",
    ]
    result = procrunner.run(scale_command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "scaled.refl").is_file()
    assert (tmp_path / "scaled.expt").is_file()
    assert (tmp_path / "scaled.mtz").is_file()

    i_obs = iotbx.merging_statistics.select_data(str(tmp_path / "scaled.mtz"), None)
    result = iotbx.merging_statistics.dataset_statistics(
        i_obs=i_obs,
        use_internal_variance=False,
        eliminate_sys_absent=False,
    )

    assert result.overall.cc_one_half > 0.8



from __future__ import annotations

import json
import os
from pathlib import Path

import procrunner

from dials.command_line import stereographic_projection


def test_stereographic_projection(dials_data, tmp_path):
    result = procrunner.run(
        (
            "dials.stereographic_projection",
            dials_data("centroid_test_data", pathlib=True) / "experiments.json",
            "hkl_limit=4",
            "plot.filename=proj.png",
            "json.filename=proj.json",
        ),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("projections.txt").is_file()
    assert tmp_path.joinpath("proj.png").is_file()
    assert tmp_path.joinpath("proj.json").is_file()

    d = json.loads(tmp_path.joinpath("proj.json").read_text())
    assert set(d) == {"data", "layout"}
    assert d["data"][0]["name"] == "stereographic_projections"
    assert len(d["data"][0]["x"]) == 289


def test_labels(dials_data, tmp_path):
    output_file = tmp_path / "proj.json"
    experiments = sorted(
        dials_data("multi_crystal_proteinase_k", pathlib=True).glob("experiments*.json")
    )
    args = [str(e) for e in experiments] + [
        f"plot.labels={' '.join(str(i) for i in range(len(experiments)))}",
        f"json.filename={output_file}",
        "hkl=1,0,0",
    ]

    cwd = Path.cwd()
    try:
        os.chdir(tmp_path)
        stereographic_projection.run(args)
    finally:
        os.chdir(cwd)
    d = json.loads(output_file.read_bytes())
    assert d["data"][0]["hoverinfo"] == "text"
    assert d["data"][0]["hovertext"] == [
        "0",
        "0",
        "1",
        "1",
        "2",
        "2",
        "3",
        "3",
        "4",
        "4",
        "5",
        "5",
        "6",
        "6",
        "7",
        "7",
    ]



"""
This module defines an abstract CrossValidator and an implementation of a
cross validator for dials.scale
"""

from __future__ import annotations

import itertools
from copy import deepcopy

import pkg_resources

from libtbx import phil
from libtbx.table_utils import simple_table
from scitbx.array_family import flex


class CrossValidator:
    """Abstract class defining common methods for cross validation and methods
    that must be implemented for concrete implementations"""

    # metadata needed when constructing the results table
    results_metadata = {
        "names": [],  # the names of the 'results' that should be printed in the
        # results table upon completion
        "indices_to_monitor": [],  # these indices of the above list will be
        # monitored to see which config gives the best values
        "best_criterion": [],  # define 'best' for the above indices ('min' or 'max')
    }

    def __init__(self, experiments, reflections):
        self.experiments = experiments
        self.reflections = reflections
        self.results_dict = {}

    def run_script(self, params, config_no):
        """Run the appropriate command line script with the params, get the
        free/work set results and add to the results dict. Indicate the
        configuration number being run."""
        raise NotImplementedError()

    def get_results_from_script(self, script):
        """Return the work/free results list from the command line script object"""
        raise NotImplementedError()

    def get_parameter_type(self, name):
        """Find the parameter type for a discreet phil option - bool or choice."""
        raise NotImplementedError()

    def set_parameter(self, params, name, val):
        """Find the name in the params scope extract and set it to the val"""
        raise NotImplementedError()

    def set_free_set_offset(self, params, n):
        """Set the free set offset in the correct place in the scope"""
        raise NotImplementedError()

    def get_free_set_percentage(self, params):
        """Inspect the free set percentage in the correct place in the scope"""
        raise NotImplementedError()

    @staticmethod
    def _avg_sd_from_list(lst):
        """simple function to get average and standard deviation"""
        arr = flex.double(lst)
        avg = round(flex.mean(arr), 5)
        std = round(arr.standard_deviation_of_the_sample(), 5)
        return avg, std

    def create_results_dict(self, n_options):
        """Create a results dict of the correct size for filling later"""
        if n_options == 1:
            self.results_dict[0] = {"configuration": ["user"]}
            for name in self.results_metadata["names"]:
                self.results_dict[0][name] = []
        else:
            self.results_dict = {}
            for i in range(n_options):
                self.results_dict[i] = {"configuration": []}
                for name in self.results_metadata["names"]:
                    self.results_dict[i][name] = []

    def set_results_dict_configuration(self, keys, values):
        """Add configuration information to the results dict"""
        assert len(keys) == len(values)
        for i, v in enumerate(itertools.product(*values)):
            e = dict(zip(keys, v))
            for k, val in e.items():
                self.results_dict[i]["configuration"].append(str(k) + "=" + str(val))

    def add_results_to_results_dict(self, config_no, results):
        """Add the results to the correct place in the dict"""
        assert len(results) == len(self.results_metadata["names"])
        for name, result in zip(self.results_metadata["names"], results):
            self.results_dict[config_no][name].append(result)

    def interpret_results(self):
        """Inspect the data in results_dict, make a nice table with the mean and
        average over many attempts and indicate the 'best' option"""
        rows = []
        headers = ["option", ""] + self.results_metadata["names"]
        monitored_values = []

        # Construct the rows, using the metadata from the results dict
        for v in self.results_dict.values():
            config_str = " ".join(v["configuration"])
            vals, stds = [], []
            for i, name in enumerate(self.results_metadata["names"]):
                val, std = self._avg_sd_from_list(v[name])
                vals.append(val)
                stds.append(std)
                if i in self.results_metadata["indices_to_monitor"]:
                    monitored_values.append(val)
            rows.append([config_str, "mean"] + [str(i) for i in vals])
            rows.append(["", "std dev"] + [str(i) for i in stds])

        # Now go through monitored values, finding the best and adding a '*'
        n_monitored = len(self.results_metadata["indices_to_monitor"])
        for i in range(n_monitored):
            vals = monitored_values[i::n_monitored]
            if self.results_metadata["best_criterion"][i] == "max":
                best_idx = vals.index(max(vals)) * 2  # *2 to skip std rows
            elif self.results_metadata["best_criterion"][i] == "min":
                best_idx = vals.index(min(vals)) * 2  # *2 to skip std rows
            rows[best_idx][self.results_metadata["indices_to_monitor"][i] + 2] += "*"
            # line above, 2 is to offset first two columns in table

        return simple_table(rows, headers)


class DialsScaleCrossValidator(CrossValidator):
    """An implementation of the CrossValidator for running dials.scale"""

    results_metadata = {  # metadata used when constructing the results table
        "names": [
            "work Rmeas",
            "free Rmeas",
            "Rmeas gap",
            "work CC1/2",
            "free CC1/2",
            "CC1/2 gap",
        ],
        "indices_to_monitor": [1, 2, 4, 5],  # these indices of the above list
        # will be monitored to see which config gives the best values
        "best_criterion": ["min", "min", "max", "min"],
    }

    def get_results_from_script(self, script):
        """Return the work/free results list from the command line script object"""
        result = script.scaler.work_free_stats
        return result

    def get_parameter_type(self, name):
        """Find the parameter type for a scaling phil option."""
        phil_scope = phil.parse(
            """include scope dials.command_line.scale.phil_scope""",
            process_includes=True,
        )
        obj = phil.find_scope(phil_scope, name)
        if not obj:
            raise ValueError(
                """Unable to resolve %s in the phil scope, make sure full phil path
is provided. For example, physical.decay_correction rather than decay_correction"""
                % name
            )
        return obj.type.phil_type  # a str: "int", "bool" etc

    def set_parameter(self, params, name, val):
        """Find the name in the params scope extract and set it to the val"""
        if name == "model":
            params.model = val
            return params
        available_models = [
            entry_point.name
            for entry_point in pkg_resources.iter_entry_points(
                "dxtbx.scaling_model_ext"
            )
        ]
        phil_branches = [
            params.weighting.error_model,
            params.cut_data,
            params.scaling_options,
            params.reflection_selection,
            params.reflection_selection.random,
            params.reflection_selection.random.multi_dataset,
        ]
        if params.model:
            phil_branches.append(params.__getattribute__(str(params.model)))
        elif ("." in name) and (name.split(".")[0] in available_models):
            # if the user hasn't specified the model, but have done
            # e.g physical.parameter = *, then set model=physical
            params.model = name.split(".")[0]
            phil_branches.append(params.__getattribute__(str(params.model)))
        if "." in name:  # handle e.g physical.absorption_correction
            name = name.split(".")[-1]
        for branch in phil_branches:
            try:
                branch.__setattr__(name, val)
                return params
            except AttributeError:
                pass
        # if get here, haven't found what we're trying to set
        raise ValueError("Unable to set chosen attribute " + str(name) + "=" + str(val))

    def set_free_set_offset(self, params, n):
        """Set the free set offset in the correct place in the scope"""
        params.scaling_options.free_set_offset = n
        return params

    def get_free_set_percentage(self, params):
        """Inspect the free set percentage in the correct place in the scope"""
        return params.scaling_options.free_set_percentage

    def run_script(self, params, config_no):
        """Run the scaling script with the params, get the free/work set results
        and add to the results dict"""
        from dials.algorithms.scaling.algorithm import ScalingAlgorithm

        params.scaling_options.__setattr__("use_free_set", True)
        algorithm = ScalingAlgorithm(
            params,
            experiments=deepcopy(self.experiments),
            reflections=deepcopy(self.reflections),
        )
        algorithm.run()
        results = self.get_results_from_script(algorithm)
        self.add_results_to_results_dict(config_no, results)



from __future__ import annotations

import logging
import os

from dxtbx.serialize import xds
from iotbx.xds import spot_xds
from scitbx import matrix

logger = logging.getLogger(__name__)


def dump(experiments, reflections, directory):
    """Dump the files in XDS format"""
    if len(experiments) > 0:
        for i, experiment in enumerate(experiments):
            suffix = ""
            if len(experiments) > 1:
                suffix = "_%i" % (i + 1)

            sub_dir = f"{directory}{suffix}"
            if not os.path.isdir(sub_dir):
                os.makedirs(sub_dir)
            # XXX imageset is getting the experimental geometry from the image files
            # rather than the input models.expt file
            imageset = experiment.imageset
            imageset.set_detector(experiment.detector)
            imageset.set_beam(experiment.beam)
            imageset.set_goniometer(experiment.goniometer)
            imageset.set_scan(experiment.scan)

            if experiment.crystal is None:
                space_group_number = None
                real_space_a = None
                real_space_b = None
                real_space_c = None
                job_card = "XYCORR INIT COLSPOT IDXREF DEFPIX INTEGRATE CORRECT"
            else:
                crystal_model = experiment.crystal
                crystal_model = crystal_model.change_basis(
                    crystal_model.get_space_group()
                    .info()
                    .change_of_basis_op_to_reference_setting()
                )
                space_group_number = crystal_model.get_space_group().type().number()
                A = matrix.sqr(crystal_model.get_A())
                A_inv = A.inverse()
                real_space_a = A_inv.elems[:3]
                real_space_b = A_inv.elems[3:6]
                real_space_c = A_inv.elems[6:9]
                job_card = ("XYCORR INIT DEFPIX INTEGRATE CORRECT",)

            to_xds = xds.to_xds(imageset)
            xds_inp = os.path.join(sub_dir, "XDS.INP")
            xparm_xds = os.path.join(sub_dir, "XPARM.XDS")
            logger.info("Exporting experiment to %s", xds_inp)
            with open(xds_inp, "w") as f:
                f.write(
                    to_xds.XDS_INP(
                        space_group_number=space_group_number,
                        real_space_a=real_space_a,
                        real_space_b=real_space_b,
                        real_space_c=real_space_c,
                        job_card=job_card,
                    )
                )
            if space_group_number:
                logger.info("Exporting crystal model to %s", xparm_xds)
                with open(xparm_xds, "w") as f:
                    f.write(
                        to_xds.xparm_xds(
                            real_space_a, real_space_b, real_space_c, space_group_number
                        )
                    )

            if reflections is not None and len(reflections) > 0:
                ref_cryst = reflections.select(reflections["id"] == i)
                export_spot_xds(ref_cryst, os.path.join(sub_dir, "SPOT.XDS"))

    else:
        if not os.path.isdir(directory):
            os.makedirs(directory)
        export_spot_xds(reflections, os.path.join(directory, "SPOT.XDS"))


def export_spot_xds(reflections, filename):
    if reflections is not None and len(reflections) > 0:
        centroids = reflections["xyzobs.px.value"]
        intensities = reflections["intensity.sum.value"]
        miller_indices = None
        if "miller_index" in reflections:
            miller_indices = reflections["miller_index"]
            selection = miller_indices != (0, 0, 0)
            miller_indices = miller_indices.select(selection)
            if len(miller_indices) == 0:
                miller_indices = None
            else:
                centroids = centroids.select(selection)
                intensities = intensities.select(selection)
        xds_writer = spot_xds.writer(
            centroids=centroids, intensities=intensities, miller_indices=miller_indices
        )
        logger.info("Exporting spot list as %s", filename)
        xds_writer.write_file(filename=filename)



from __future__ import annotations

import math
import pickle
import sys
from concurrent.futures import ProcessPoolExecutor, as_completed

import iotbx.phil
from scitbx.array_family import flex

import dials.util
from dials.algorithms.spot_finding.factory import SpotFinderFactory
from dials.algorithms.spot_finding.factory import phil_scope as spot_phil
from dials.util.options import ArgumentParser, flatten_experiments

help_message = """

Identify unreliable pixels in a sequence of images: defined to be those pixels
which read as signal in spot finding in 50% or over of images. Outputs data in
the order slow, fast (i.e. row-major order) as ::

  mask[slow, fast] = 16

Assuming that these will be used to assign values to e.g. a numpy array. 16 in
this case indicates "noisy" (e.g. bit 4 set to high:
https://manual.nexusformat.org/classes/applications/NXmx.html)

Examples::

  dials.find_bad_pixels data_master.h5 [nproc=8]

"""

phil_scope = iotbx.phil.parse(
    """
images = None
  .type = ints
  .help = "Images on which to perform the analysis (otherwise use all images)"
image_range = None
  .type = ints(value_min=0, size=2)
  .help = "Image range for analysis e.g. 1,1800"
nproc = 1
  .type = int(value_min=1)
    .help = "The number of processes to use."

output {
    mask = pixels.mask
        .type = path
        .help = "Output mask file name"
    print_values = False
        .type = bool
        .help = "Print bad pixel values"
}
"""
)


def find_constant_signal_pixels(imageset, images):
    """Find pixels which are constantly reporting as signal through the
    images in imageset: on every image the pixel dispersion index is computed,
    and signal pixels identified using the default settings. A map is then
    calculated of the number of times a pixel is identified as signal: if this
    is >= 50% of the images (say) that pixel is untrustworthy."""

    panels = imageset.get_detector()

    # only cope with monilithic detectors or the I23 Pilatus 12M
    assert len(panels) in (1, 24)

    # trusted range the same for all panels anyway
    detector = panels[0]
    trusted = detector.get_trusted_range()

    # construct an integer array same shape as image; accumulate number of
    # "signal" pixels in each pixel across data

    total = None

    for idx in images:
        pixels = imageset.get_raw_data(idx - 1)
        known_mask = imageset.get_mask(idx - 1)

        # apply known mask
        for _pixel, _panel, _mask in zip(pixels, panels, known_mask):
            _pixel.set_selected(~_mask, -1)
            for f0, s0, f1, s1 in _panel.get_mask():
                blank = flex.int(flex.grid(s1 - s0, f1 - f0), 0)
                _pixel.matrix_paste_block_in_place(blank, s0, f0)

        if len(pixels) == 1:
            data = pixels[0]
        else:
            ny, nx = pixels[0].focus()
            data = flex.int(flex.grid(24 * ny + 23 * 17, nx), -1)
            for j in range(24):
                data.matrix_paste_block_in_place(pixels[j], j * (ny + 17), 0)

        negative = data < int(round(trusted[0]))
        hot = data > int(round(trusted[1]))
        bad = negative | hot

        data = data.as_double()

        spot_params = spot_phil.fetch(
            source=iotbx.phil.parse("min_spot_size=1")
        ).extract()
        threshold_function = SpotFinderFactory.configure_threshold(spot_params)
        peak_pixels = threshold_function.compute_threshold(data, ~bad)

        if total is None:
            total = peak_pixels.as_1d().as_int()
        else:
            total += peak_pixels.as_1d().as_int()

    return total


@dials.util.show_mail_handle_errors()
def run(args=None):
    usage = "dials.find_bad_pixels [options] (data_master.h5|data_*.cbf)"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        read_experiments_from_images=True,
        epilog=help_message,
    )

    params, options = parser.parse_args(args, show_diff_phil=True)

    experiments = flatten_experiments(params.input.experiments)

    if len(experiments) == 0:
        parser.print_help()
        sys.exit("Please pass an experiment list or image data\n")
        return

    # assume that all data come from the same detector, if multiple experiments
    # are passed (though could use the same principle for finding the set of
    # all pixels which _can_ be unreliable too...)

    hot_mask = None

    for experiment in experiments:
        imageset = experiment.imageset

        # TODO fix this for > 1 panel detectors
        detector = experiment.detector[0]
        nfast, nslow = detector.get_image_size()

        first, last = experiment.scan.get_image_range()
        images = range(first, last + 1)

        # TODO tidy up the logic here - it is _correct_ but a bit ...
        # convoluted
        if params.images is None and params.image_range is not None:
            start, end = params.image_range
            params.images = list(range(start, end + 1))

        if params.images:
            if min(params.images) < first or max(params.images) > last:
                sys.exit("Image outside of scan range")
            images = params.images

        # work around dxtbx "features" to do with counting from (0, 1, -1, n)
        images = [i - first + 1 for i in images]

        n = int(math.ceil(len(images) / params.nproc))
        chunks = [images[i : i + n] for i in range(0, len(images), n)]

        assert len(images) == sum(len(chunk) for chunk in chunks)

        if len(chunks) < params.nproc:
            params.nproc = len(chunks)

        total = None
        with ProcessPoolExecutor(max_workers=params.nproc) as p:
            jobs = []
            for j in range(params.nproc):
                jobs.append(p.submit(find_constant_signal_pixels, imageset, chunks[j]))
            for job in as_completed(jobs):
                if total is None:
                    total = job.result()
                else:
                    total += job.result()

        if hot_mask is None:
            hot_mask = total >= (len(images) // 2)
        else:
            hot_mask = hot_mask & (total >= (len(images) // 2))

    hot_pixels = hot_mask.iselection()

    for h in hot_pixels:
        print(f"mask: {h // nfast} {h % nfast} 16")

    if params.output.mask:
        mask = flex.bool(
            flex.grid(reversed(imageset.get_detector()[0].get_image_size())), True
        )

        for h in hot_pixels:
            mask[h // nfast, h % nfast] = False

        with open(params.output.mask, "wb") as fh:
            pickle.dump(mask, fh, pickle.HIGHEST_PROTOCOL)

        print(f"Wrote pixel mask to {params.output.mask}")


if __name__ == "__main__":
    run()



from __future__ import annotations

import json


class ReciprocalLatticeJson:
    def __init__(self, experiments, reflections):
        self.experiments = experiments
        self.reflections = reflections
        self.reflections.centroid_px_to_mm(experiments)
        self.reflections.map_centroids_to_reciprocal_space(experiments)

    def as_dict(self, n_digits=None):
        rlp = self.reflections["rlp"]
        if n_digits is not None:
            rlp = rlp.round(n_digits)
        flat_rlp = []
        for r in rlp:
            flat_rlp.extend(r)

        if "imageset_id" in self.reflections:
            imageset_id = list(self.reflections["imageset_id"])
            expt_id = list(self.reflections["id"])
        else:
            imageset_id = list(self.reflections["id"])
            expt_id = None

        d = {"rlp": flat_rlp, "imageset_id": imageset_id, "experiment_id": expt_id}
        return d

    def as_json(self, filename=None, compact=False, n_digits=None, experiments=None):
        d = self.as_dict(n_digits=n_digits)
        if experiments:
            d["experiments"] = experiments.to_dict()
        if compact:
            text = json.dumps(d, separators=(",", ":"), ensure_ascii=True)
        else:
            text = json.dumps(d, separators=(",", ": "), indent=1, ensure_ascii=True)
        if filename is not None:
            with open(filename, "w") as f:
                f.write(text)
        else:
            return text



from __future__ import annotations


class GaussianRSProfileModelExt:
    """An extension class implementing a reciprocal space gaussian profile model."""

    name = "gaussian_rs"

    default = True

    @staticmethod
    def phil():
        from dials.algorithms.profile_model.gaussian_rs import phil_scope

        return phil_scope

    @staticmethod
    def algorithm():
        from dials.algorithms.profile_model.gaussian_rs import Model

        return Model

    @classmethod
    def from_dict(cls, d):
        return cls.algorithm().from_dict(d)
